## How to become a DDMS

* [Introduction](#introduction)
* [Register as a DDMS](#register)
* [Create a Storage schema](#create-schema)
* [Expose legal tag and ACLs through your APIs](#expose-legal)
* [Optional: Expose compliance for derivative data](#expose-derivative)
* [Ingest data and create a shadow record](#shadow-record)
* [Perform compliance and ACL checks using shadow records ](#validate)
* [Client retrieves the bulk data](#retrieve)

## Introduction <a name="introduction"></a>
A Domain Data Management Service (DDMS) can be seen as any source of truth for data that manages the data life cycle, satisfies given mandatory data access concerns, and makes its data globally discoverable and retrievable through the OSDU. It could be a standalone service dedicated to a specific data type or a subcomponent of an application or platform. It simply enables its data to be retrieved outside of its regular scope.

A DDMS needs to enforce the common concerns of 

*  Legal compliance
*  Data access authorization
*  Discovery
*  Retrieval of the data based on discovery

OSDU solves these concerns primarily using Storage records. A Storage record is metadata pertaining to the bulk data stored in the DDMS. Every record created in Storage enforces that ACLs are assigned, checks compliance and then indexes the record into search, making it discoverable.

The following is the preferred method of using Records to enable these concerns for a DDMS.

## Register as a DDMS <a name="register"></a>

The first step is to register as a DDMS. This makes your DDMS discoverable to clients and presents them with an API definition that tells them how to retrieve the bulk data when a record from their DDMS is discovered. 

The only API that needs to be defined is the one that tells them how to retrieve the bulk data based on an Id.

Note that you can register as much of your API specification as you like. You only need to define the method clients should use to retrieve the bulk data using the custom property x-ddms-retrieve-entity: true.

<details><summary>curl</summary>

```

    curl --request POST \
    --url '/api/register/v1/ddms' \
    --header 'accept: application/json' \
    --header 'authorization: Bearer <JWT>' \
    --header 'content-type: application/json' \
    --header 'data-partition-id: common' \
    --data '{
    {
    "id": "{DDMS-ID}",
    "name": "logDDMS",
    "description": "My test ddms.",
    "contactEmail": "test@test.com",
    "interfaces": [
        {
        "entityType": "wellbore",
        "schema": {
            "openapi": "3.0.0",
            "info": {
            "description": "This is a sample Wellbore domain DM service.",
            "version": "1.0.0",
            "title": "OSDU Wellbore Domain DM Service",
            "contact": {
                "email": "osdu-sre@opengroup.org"
            }
            },
            "servers": [
            {
                "url": "https://subsurface.data.osdu.com/v1"
            }
            ],
            "tags": [
            {
                "name": "wellbore",
                "description": "Wellbore data type services"
            }
            ],
            "paths": {
            "/wellbore/{wellboreId}": {
                "get": {
                "tags": [
                    "wellbore"
                ],
                "summary": "Find wellbore by ID",
                "description": "Returns a single wellbore",
                "operationId": "getWellboreById",
                "x-ddms-retrieve-entity": true,
                "parameters": [
                    {
                    "name": "wellboreId",
                    "in": "path",
                    "description": "ID of wellbore to return",
                    "required": true,
                    "schema": {
                        "type": "string"
                    }
                    }
                ],
                "responses": {
                    "200": {
                    "description": "successful operation",
                    "content": {
                        "application/json": {
                        "schema": {
                            "$ref": "#/components/schemas/wellbore"
                        }
                        }
                    }
                    },
                    "400": {
                    "description": "Invalid ID supplied"
                    },
                    "401": {
                    "description": "Not authorized"
                    },
                    "404": {
                    "description": "Wellbore not found"
                    }
                }
            }
        }
    }

```

</details>

## Create a Storage schema <a name="create-schema"></a>

It is up to the bulk data store to determine what properties of the bulk data they want to push into a Storage record and to make discoverable within the OSDU.

They define a storage schema to represent this. The schema is a list of properties and the type of data they represent that will be on the Record.

When deploying your service, you should do the one-time operation of publishing the schema via the Storage APIs. 

<details><summary>curl</summary>

```

    curl --request POST \
    --url '/api/storage/v2/schemas' \
    --header 'accept: application/json' \
    --header 'authorization: Bearer <JWT>' \
    --header 'content-type: application/json' \
    --header 'data-partition-id: common' \
    --data '{
        "kind": "common:welldb:wellbore:1.0.0",
        "schema": [
            {
            "path": "name",
            "kind": "string"
            },
            {
            "path": "ddmsId",
            "kind": "string"
            },
            {
            "path": "localId",
            "kind": "string"
            },
            {
            "path": "entityType",
            "kind": "string"
            }]
    }'

```

</details>

This will then allow any Record that references this schema to be indexed in the DE search. Without this, the Record will be published but without any of the data and it will be hidden by default in search.

Notice, we are also declaring 3 properties:

**ddmsId** is the id used when you register as a DDMS.

**entityType** is the domain object type the data represents, e.g. ‘seismic’, ‘well’.

**localId** is the id of the bulk data as it is referenced in your DDMS. The end user should be able to use this id to retrieve the bulk data from your APIs.

These act as well-known properties that should be added to the record by your DDMS. Clients can then use this information to retrieve the bulk data after discovery using the DDMS registration APIs. Every schema created should declare these properties to use this pattern of ingestion.

## Expose legal tag and ACLs through your APIs <a name="expose-legal"></a>
Unless you have a scenario where you know what legal tag and ACL should be applied to the data you are ingesting, you will need to expose the legal tag and ACL in your ingestion APIs. This allows your clients to supply the legal tag and ACL themselves.

You can expose the same interface as the Storage records API, allowing you to assign them to the record you create.

<details>

```

    "acl": {
      "viewers": ['data.default.viewers@{datapartition}.{domain}.com'],
      "owners": ['data.default.owners@{datapartition}.{domain}.com']
    },
    "legal": {
      "legaltags": ['common-sample-legaltag']

```
</details>

## Optionally expose derivative compliance through your APIs <a name="expose-derivative"></a>
If you expect derivative data to be stored in your DDMS, you need to expose 2 more properties through your APIs that can be appended to your Storage record.

Again, you can expose the same interface as the Storage records API, allowing you to assign them to the record directly. The 2 properties are:

* otherRelevantDataCountries: The alpha 2 country code of the country where the derivative was created or calculated
* parents: The record ids and versions of the Records this derivative was created from

If a derivative is being created then a legal tag does not need to be assigned as it inherits this from its parents.

<details><summary>curl</summary>

```

    "legal" :{
                  "otherRelevantDataCountries": ["US"] 
      },
      "ancestry" :{
                  "parents": ["common:id:1:version", "common:id:2:version"]
      }    

```
</details>

## Ingest data and create a shadow record <a name="shadow-record"></a>

Whenever bulk data is ingested, you need to create a shadow record within Storage.  This shadow record represents the specific bulk data instance in a 1:1 relationship and makes each instance globally discoverable.

When you create the shadow record using the Storage API, forward on the original callers jwt token.

First, you should store the bulk data, and then create the shadow Record. This way, a global piece of data is not discoverable before the bulk data is available. If this is not successful, e.g. because an invalid legal tag is provided, the request will fail and you should return this response to the client and attempt to clean up the bulk data.

Remember, you should append your DDMS Id, entityType and the bulk data’s local id to the Storage record.

<details><summary>curl</summary>

```

    curl --request PUT \
    --url '/api/storage/v2/records' \
    --header 'accept: application/json' \
    --header 'authorization: Bearer <JWT>' \
    --header 'content-type: application/json' \
    --header 'data-partition-id: common' \
    --data '[
    {
        "kind": "common:welldb:wellbore:1.0.0",
        "acl": {
        "viewers": ['data.default.viewers@{datapartition}.{domain}.com'],
        "owners": ['data.default.owners@{datapartition}.{domain}.com']
        },
        "legal": {
        "legaltags": ['common-sample-legaltag'],
        "otherRelevantDataCountries": ["FR”]
        },
        "data": {
        "name": "well1",
        "entityType": wellbore,
        "ddmsId": "abcdef",
        "localId": "123456"
        }]'

```

</details>


## Perform compliance and ACL checks using shadow records <a name="validate"></a>

As mentioned, a DDMS should create a shadow record for every instance of bulk data ingested into their data store. This can have advantages beyond global discover-ability. Whenever you request a storage record, both compliance and entitlements are checked before returning the data. A DDMS can use this to their advantage.

By forwarding on any request by the client to retrieve the record, you can delegate these responsibilities to the Storage service. If OSDU returns the Record, the client can access both this and the bulk data, and so you can return the same to the client or only the Record.

<details><summary>curl</summary>

```

    curl --request POST \
    --url '/api/storage/v2/query/records:batch' \
    --header 'authorization: Bearer <JWT>' \
    --header 'content-type: application/json' \
    --header 'data-partition-id: common' \
    --header 'frame-of-reference: NONE' \
    --data '{
        "records": [
            "common:test:fetchtest-1",
            "common:test:fetchtest-2",
            "common:test:fetchtest-4",
            "common:test:fetchtest-5",
            "common:test:fetchtest-6"

        ]
    }

```

</details>

In this scenario, you also don’t need to store the ACL or legal tag information in your DDMS because those are being retrieved directly from the OSDU in this request. However, you need to either store or be able to generate the Storage record ID needed to retrieve the record for the bulk data requested.

## Client retrieves the bulk data <a name="retrieve"></a>

Imagine the client discovered a record with the following data

<details><summary>curl</summary>

```

    "data": {
      "name": "well1",
      "entityType": wellbore,
      "ddmsId": "abcdef",
      "localId": "123456"
    }

```

</details>

They can use the ddmsId property of the data object to retrieve the API definition of the DDMS you registered at the start. 

<details><summary>curl</summary>

```

    curl --request GET \
    --url '/api/register/v1/ddms/abcdef' \
    --header 'authorization: Bearer <JWT>' \
    --header 'content-type: application/json' \
    --header 'data-partition-id: common' 

```
</details>
    
This will return them the registered DDMS with the API specification. So returning to the original one we registered would look like

<details>

```
    {
        "entityType": "wellbore",
        "schema": {
            "openapi": "3.0.0",
            "info": {
            "description": "This is a sample Wellbore domain DM service.",
            "version": "1.0.0",
            "title": "OSDU Wellbore Domain DM Service",
            "contact": {
                "email": "osdu-sre@opengroup.org"
            }
            },
            "servers": [
            {
                "url": "https://subsurface.data.osdu.com/v1"
            }
            ],
            "tags": [
            {
                "name": "wellbore",
                "description": "Wellbore data type services"
            }
            ],
            "paths": {
            "/wellbore/{wellboreId}": {
                "get": {
                "tags": [
                    "wellbore"
                ],
                "summary": "Find wellbore by ID",
                "description": "Returns a single wellbore",
                "operationId": "getWellboreById",
                "x-ddms-retrieve-entity": true,
                "parameters": [
                    {
                    "name": "wellboreId",
                    "in": "path",
                    "description": "ID of wellbore to return",
                    "required": true,
                    "schema": {
                        "type": "string"
                    }
                    }
                ],
                "responses": {
                    "200": {
                    "description": "successful operation",
                    "content": {
                        "application/json": {
                        "schema": {
                            "$ref": "#/components/schemas/wellbore"
                        }
                        }
                    }
                    },
                    "400": {
                    "description": "Invalid ID supplied"
                    },
                    "401": {
                    "description": "Not authorized"
                    },
                    "404": {
                    "description": "Wellbore not found"
                    }
                }
            }
        }

```
</details>

They can then use the entityType property to work out which API definition to use, and then use the localId property to work out how to create the API call to retrieve the bulk data from the DDMS using the API defined.

Currently, this is a manual step for the user and so they have to understand your API definition. However once they do, they can apply the same pattern to any Record discovered for your DDMS.

Using the returned specification and data we discovered, the resulting API call the user would be expected to make to retrieve the bulk data would be

<details><summary>curl</summary>

```
    curl --request GET \
    --url 'https://subsurface.data.osdu.com/v1/wellbore/123456' \
    --header 'authorization: Bearer <JWT>' \
    --header 'content-type: application/json' \
    --header 'data-partition-id: common' 

```

</details>

Where 123456 is the localId stored in the records data and the URL is defined in the API spec in the server, and path sections of the operation that has the property x-ddms-retrieve-entity. 
